#include "stdafx.h"
#include "Arcball.h"            

#include <gl/gl.h>
// Arcball sphere constants:
// Diameter:     2.0f
// Radius:       1.0f


//Create
Arcball::Arcball()
{
    // Clear initial values
	vecClick[0] = vecClick[1] = vecClick[2] = 0.0f;
	vecDrag[0]  = vecDrag[1]  = vecDrag[2]  = 0.0f;
	
    // Set initial mouse bounds for arcball
	reshape(1024, 768);
	
 	// ʼɵλ
 	for (int i=0; i<16; i++)
 		transform.M[i] = 0;
 	transform.M[0] = transform.M[5] = transform.M[10] = transform.M[15] = 1.0f;
 
 	for (int i=0; i<8; i++)
 		lastRot.M[i] = 0.0f;
 	lastRot.M[0] = lastRot.M[4] = lastRot.M[8] = 1.0f;
 
 	for (int i=0; i<8; i++)
 		thisRot.M[i] = 0.0f;
 	thisRot.M[0] = thisRot.M[4] = thisRot.M[8] = 1.0f;
}

Arcball::~Arcball()
{

}

//Set mouse bounds
void Arcball::reshape(float width, float height)
{
	assert((width > 1.0f) && (height > 1.0f));
	
	//Set adjustment factor for width and height
	mbWidth  = 1.0f / ((width  - 1.0f) * 0.5f);
	mbHeight = 1.0f / ((height - 1.0f) * 0.5f);
}

	
// Starting rotation. Call when drag-mouse down.
// parameter "ptClicked" is the mouse coordinates when mouse clicked,
void Arcball::mouse_down(float x, float y)
{
	// Set Last Static Rotation To Last Dynamic One
	lastRot = thisRot;
	// Update Start Vector And Prepare For Dragging
	// Map the point to the sphere
	map_to_sphere(x, y, vecClick);
}


// Drag and move the mouse. Call when drag-mouse moves to rotate.
void Arcball::mouse_move(float x, float y)
{
	//Map the point to the sphere
	map_to_sphere(x, y, vecDrag);
	
	// Update End Vector And Get Rotation As Quaternion
	// Mouse drag, calculate rotation
	// Return the quaternion equivalent to the rotation
	Quat4 ThisQuat;	

	//Compute the vector perpendicular to the begin and end vectors
	float perp[3] = {
		vecClick[1] * vecDrag[2] - vecClick[2] * vecDrag[1], 
		vecClick[2] * vecDrag[0] - vecClick[0] * vecDrag[2],
		vecClick[0] * vecDrag[1] - vecClick[1] * vecDrag[0]
	};

	float perpLen = sqrt(perp[0] * perp[0] + perp[1] * perp[1] + perp[2] * perp[2]);	
	if (perpLen > EPSILON)  // If the length of the perpendicular vector is non-zero
	{
		//We're ok, so return the perpendicular vector as the transform after all
		ThisQuat.s.X = perp[0];
		ThisQuat.s.Y = perp[1];
		ThisQuat.s.Z = perp[2];
		//In the quaternion values, w is cosine (theta / 2), where theta is rotation angle
		ThisQuat.s.W = vecClick[0] * vecDrag[0] + vecClick[1] * vecDrag[1] + vecClick[2] * vecDrag[2];
	}
	else  // If is zero
	{
		// The begin and end vectors coincide, so return an identity transform
		ThisQuat.s.X =
		ThisQuat.s.Y = 
		ThisQuat.s.Z = 
		ThisQuat.s.W = 0.0f;
	}
	
	Matrix3fSetRotationFromQuat4f(&thisRot, &ThisQuat);		// Convert Quaternion Into Matrix3fT
	Matrix3fMulMatrix3f(&thisRot, &lastRot);				// Accumulate Last Rotation Into This One
	Matrix4fSetRotationFromMatrix3f(&transform, &thisRot);	// Set Our Final Transform's Rotation From This One
}


// Apply Dynamic Transform
void Arcball::apply_rotation()
{
	glMultMatrixf(transform.M);
}


// Reset Rotation
void Arcball::reset_rotation()   	
{
	Matrix3fSetIdentity(&lastRot);							
	Matrix3fSetIdentity(&thisRot);							
	Matrix4fSetRotationFromMatrix3f(&transform, &thisRot);	
}

void Arcball::map_to_sphere(float ptMouseX, float ptMouseY, float* vecNew) const
{ 	
    // Adjust point coords and scale down to range of [-1 ... 1]
    ptMouseX  = (ptMouseX * mbWidth)  - 1.0f;
    ptMouseY  = 1.0f - (ptMouseY * mbHeight);

    // Compute the square of the length of the vector to the point from the center
	float length = (ptMouseX * ptMouseX) + (ptMouseY * ptMouseY);	
 	
    // If the point is mapped outside of the sphere... (length > radius squared)
    if (length > 1.0f)
    {
        // Compute a normalizing factor (radius / sqrt(length))
        float norm = 1.0f / sqrt(length);
		
        // Return the "normalized" vector, a point on the sphere
		vecNew[0] = ptMouseX * norm;
        vecNew[1] = ptMouseY * norm;
        vecNew[2] = 0.0f;
    }
    else    // Else it's on the inside
    {
        // Return a vector to a point mapped inside the sphere sqrt(radius squared - length)
		vecNew[0] = ptMouseX;
		vecNew[1] = ptMouseY;
		vecNew[2] = sqrt(1.0f - length);
    }
}


void Arcball::Copy(Arcball* arcBall) const
{
	arcBall->mbWidth = mbWidth;
	arcBall->mbHeight = mbHeight;

	arcBall->vecClick[0] = vecClick[0];
	arcBall->vecClick[1] = vecClick[1];
	arcBall->vecClick[2] = vecClick[2];

	arcBall->vecDrag[0] = vecDrag[0];
	arcBall->vecDrag[1] = vecDrag[1];
	arcBall->vecDrag[2] = vecDrag[2];

	for (int i=0; i<9; i++)
		arcBall->lastRot.M[i] = lastRot.M[i];
	for (int i=0; i<9; i++)
		arcBall->thisRot.M[i] = thisRot.M[i];
	for (int i=0; i<15; i++)
		arcBall->transform.M[i] = transform.M[i];
}

void Arcball::Set(const Arcball* arcBall)
{
	mbWidth = arcBall->mbWidth;
	mbHeight = arcBall->mbHeight;

	vecClick[0] = arcBall->vecClick[0];
	vecClick[1] = arcBall->vecClick[1];
	vecClick[2] = arcBall->vecClick[2];

	vecDrag[0] = arcBall->vecDrag[0];
	vecDrag[1] = arcBall->vecDrag[1];
	vecDrag[2] = arcBall->vecDrag[2];

	for (int i=0; i<9; i++)
		lastRot.M[i] = arcBall->lastRot.M[i];
	for (int i=0; i<9; i++)
		thisRot.M[i] = arcBall->thisRot.M[i];
	for (int i=0; i<15; i++)
		transform.M[i] = arcBall->transform.M[i];
}
