// OpenGLView.cpp : implementation file
//

#include "stdafx.h"
#include "Arcball.h"
#include "OpenGLView.h"

#include <cmath>
#include <gl/glut.h>


// COpenGLView
IMPLEMENT_DYNCREATE(COpenGLView, CView)

COpenGLView::COpenGLView()
{
	m_hRC = NULL;

	m_pArcball = new Arcball;
	InitGeometry();
}

COpenGLView::~COpenGLView()
{
	delete m_pArcball;
}

BEGIN_MESSAGE_MAP(COpenGLView, CView)
	//{{AFX_MSG_MAP(COpenGLView)
	ON_WM_CREATE()
	ON_WM_SIZE()
	ON_WM_DESTROY()	
	ON_WM_ERASEBKGND()
	ON_WM_LBUTTONDOWN()
	ON_WM_LBUTTONUP()
	ON_WM_MOUSEMOVE()
	ON_WM_RBUTTONDOWN()
	ON_WM_RBUTTONUP()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()


// COpenGLView drawing
void COpenGLView::OnDraw(CDC* pDC)
{
	// The following line is needed if you have more than one rendering context
	::wglMakeCurrent(pDC->m_hDC, m_hRC);

	// Clear the color and depth buffers.
	::glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	::glPushMatrix();	
		::glTranslatef(m_fTransX, m_fTransY, m_fTransZ); 
		::glScalef(m_fScale, m_fScale, m_fScale);	     	
		::glColor3ub(255, 255, 255);                     

		ArcballRotation();

		::glGetIntegerv(GL_VIEWPORT, m_viewPort);
		::glGetDoublev(GL_MODELVIEW_MATRIX, m_modelViewMat);
		::glGetDoublev(GL_PROJECTION_MATRIX, m_projectionMat);

		RenderSceneScalable();
	::glPopMatrix();	

	::glFlush();  // finish drawing
	::SwapBuffers(pDC->m_hDC /*wglGetCurrentDC()*/);

	// The following line is needed if you have more than one rendering context
	::ReleaseDC(GetSafeHwnd(), pDC->m_hDC);
}

void COpenGLView::RenderSceneScalable()
{
	::glutSolidSphere(0.7f, 20, 20);
}


void COpenGLView::EnableLighting(bool b)
{
	m_bEnableLighting = b;
	if (b)
		::glEnable(GL_LIGHTING);
	else
		::glDisable(GL_LIGHTING);
	
	InvalidateRect(NULL, FALSE);
}


/////////////////////////////////////////////////////////////////////////////
// COpenGLView message handlers
int  COpenGLView::OnCreate(LPCREATESTRUCT lpCreateStruct) 
{
	// sets up the drawing environment and do
	// one time initialization of OpenGL-settings

	if (CView::OnCreate(lpCreateStruct) == -1)
		return -1;

	HWND hWnd = GetSafeHwnd();
	HDC  hDC = ::GetDC(hWnd);

	// Set up the drawing environment
	if (SetWindowPixelFormat(hDC) == false)
		return -1;
	if (CreateOpenGLContext(hDC) == false)
		return -1;

	// do one-time initializing
	InitRenderContext();

	// release
	::ReleaseDC(hWnd,hDC);

	return 0;
}


void COpenGLView::OnDestroy() 
{	
	CView::OnDestroy();	
	// TODO: Add your message handler code here
	if(::wglGetCurrentContext() != NULL)
		::wglMakeCurrent(NULL,NULL);

	if(m_hRC != NULL)
	{
		::wglDeleteContext(m_hRC);
		m_hRC = NULL;
	}
}


void COpenGLView::Resize(int cx, int cy) 
{
	// We will:-O work on the projection matrix, so set matrix mode
	::glMatrixMode(GL_PROJECTION);
	::glLoadIdentity();

	// set up your viewing volume
	// NOTE: want x/y-ratio to match view ports x/y-ratio
	float aspect = (cy == 0) ? cx : cx / (float)cy;	

	if (cx > cy)
		::glOrtho(m_fOrthoLeft * aspect, m_fOrthoRight * aspect,
		          m_fOrthoBottom,        m_fOrthoTop,
		          m_fOrthoNear,          m_fOrthoFar);	
	else
		::glOrtho(m_fOrthoLeft,            m_fOrthoRight,
				  m_fOrthoBottom / aspect, m_fOrthoTop / aspect,
				  m_fOrthoNear,            m_fOrthoFar);	

	//::gluPerspective(45.0, aspect, 0.001f, 1000.0/*m_fOrthoNear, m_fOrthoFar*/);

	// use the whole client area
	::glViewport(0, 0, cx, cy);

	// set back matrix mode
	::glMatrixMode(GL_MODELVIEW);	
	::glLoadIdentity();
}


void COpenGLView::OnSize(UINT nType, int cx, int cy) 
{	
	CView::OnSize(nType, cx, cy);
	// TODO: Add your message handler code here

	// Called whenever the size of the view changes
	// We must regulate the view port in the rendering
	// context accordingly. Here we assume that the 
	// one and only rendering context is selected
	// If it for some reason is not, this must be
	// done before we do any OpenGL-action	

	Resize(cx, cy);

	m_pArcball->reshape(max(100, cx), max(100, cy));
}


BOOL COpenGLView::OnEraseBkgnd(CDC* pDC) 
{
	// TODO: Add your message handler code here and/or call default

	// To avoid flicker. This assures that
	// the client area is not blanked before a redraw starts
	// OK as long as we draw in the whole client area
	return TRUE;
}


BOOL COpenGLView::PreCreateWindow(CREATESTRUCT& cs) 
{
	// TODO: Add your specialized code here and/or call the base class
	// all OpenGL windows must have this window-style

	cs.style |= WS_CLIPSIBLINGS | WS_CLIPCHILDREN;
	return CView::PreCreateWindow(cs);
}


void COpenGLView::InitGeometry()
{
	m_bEnableLighting = true;
	m_bEnableBothSidesLight = true;
	m_bEnableAntiAliasing = true;

	m_bLButtonDown = false;
	m_bRButtonDown = false;

	m_fTransX = 0.0f;
	m_fTransY = 0.0f;
	m_fTransZ = 0.0f;

	m_fScale = 1.0f;
	m_fSpeedScale = 1.03f;	

	m_fOrthoLeft = -1.0f;
	m_fOrthoRight = 1.0f;
	m_fOrthoBottom = -1.0f;
	m_fOrthoTop = 1.0f;
	m_fOrthoNear = -1000.0f;
	m_fOrthoFar = 1000.0f;
}

//////////////////////////////////////////////////
// default one-time initialization
void COpenGLView::InitRenderContext()
{
	// Set the attributes that probably will be constant 
	// during the rendering contexts lifetime.

	//////////////////////////////////////////////////////////////////////////
	// prepare light source

	// light0
	GLfloat ambient0[]  = {0.5f, 0.5f, 0.5f, 1.0f};
	GLfloat diffuse0[]  = {0.8f, 0.8f, 0.8f, 1.0f};
	GLfloat specular0[] = {0.5f, 0.5f, 0.5f, 1.0f};
	GLfloat position0[] = {10000.0f, 10000.0f, 50000.0f, 1.0f};
	::glLightfv(GL_LIGHT0, GL_AMBIENT, ambient0);
	::glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse0);
	::glLightfv(GL_LIGHT0, GL_SPECULAR, specular0);
	::glLightfv(GL_LIGHT0, GL_POSITION, position0);

	GLfloat global_ambient[] = { 0.3f, 0.3f, 0.3f, 1 };
	::glLightModelfv(GL_LIGHT_MODEL_AMBIENT, global_ambient);

	//////////////////////////////////////////////////////////////////////////

	//glDisable(GL_COLOR_MATERIAL);

	//float MatAmbientBack[] = {0.2f, 0.2f, 0.0f, 1.0f}; // ƫ
	float MatAmbientBack[] = {0.0f, 0.2f, 0.0f, 1.0f};  // ƫ
	::glMaterialfv(GL_BACK, GL_AMBIENT, MatAmbientBack);

	if (m_bEnableBothSidesLight)
		::glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE); 
	else
		::glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_FALSE);

	if (m_bEnableLighting)
	{
		::glEnable(GL_LIGHTING);	
		::glEnable(GL_LIGHT0);
	}

	//::glEnable(GL_CULL_FACE);
	//::glCullFace(GL_BACK);

	// flat/smooth the drawing
	::glShadeModel(GL_FLAT);  // 
	::glPolygonMode(GL_FRONT_AND_BACK, GL_FILL); 	

	if (m_bEnableAntiAliasing)
	{
		::glEnable(GL_POINT_SMOOTH);
		::glHint(GL_POINT_SMOOTH_HINT, GL_NICEST);

		::glEnable(GL_LINE_SMOOTH);
		::glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);

		::glEnable(GL_BLEND);
		::glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	}
	else
	{
		::glDisable(GL_POINT_SMOOTH);
		::glDisable(GL_LINE_SMOOTH);
		::glDisable(GL_BLEND);
	}

	// λ
	::glEnable(GL_NORMALIZE);
	::glEnable(GL_AUTO_NORMAL);

	::glDepthFunc(GL_LESS);
	//::glDepthFunc(GL_LEQUAL);
	::glEnable(GL_DEPTH_TEST);

	// set background color, default is black
	::glClearColor(0, 0, 0, 1);
	::glClearDepth(1.0f);	// Depth Buffer Setup
}


// SetWindowPixelFormat
bool COpenGLView::SetWindowPixelFormat(HDC hDC)
{
	// make sure we do only one initialization
	static bool bIsInitialized = FALSE;
	ASSERT(!bIsInitialized);
	bIsInitialized = TRUE;	

	static PIXELFORMATDESCRIPTOR pfd = 
	{
		sizeof(PIXELFORMATDESCRIPTOR),  // size of this pfd
		1,                              // version number
		PFD_DRAW_TO_WINDOW |            // support window
		PFD_SUPPORT_OPENGL |            // support OpenGL
		PFD_DOUBLEBUFFER |              // double buffered
		PFD_STEREO_DONTCARE, 
		PFD_TYPE_RGBA,                  // RGBA type
		32,                             // 32-bit color depth
		8, 16, 8, 8, 8, 0,              // color bits ignored
		0,                              // no alpha buffer
		0,                              // shift bit ignored
		64,                             // accumulation buffer
		16, 16, 16, 0,                  // accum bits ignored
		32,                             // 32-bit z-buffer
		8,                              // no stencil buffer
		0,                              // no auxiliary buffer
		PFD_MAIN_PLANE,                 // main layer
		0,                              // reserved
		0, 0, 0                         // layer masks ignored
	};

	if(pfd.dwFlags & PFD_NEED_PALETTE)
		::AfxMessageBox( _T("Colors will not display properly,\nOnly 256 colors available\nNo palette handling included") );

	int nPixelFormat = ChoosePixelFormat(hDC, &pfd);
	if (nPixelFormat == 0)  // choose default
	{
		nPixelFormat = 1;
		if (::DescribePixelFormat(hDC, nPixelFormat, sizeof(PIXELFORMATDESCRIPTOR), &pfd) == 0)
			return false;
	}

	if (!::SetPixelFormat(hDC, nPixelFormat, &pfd))
		return false;

	return true;
}


//********************************************
// CreateOpenGLContext
// Create an OpenGL rendering context
//********************************************
bool COpenGLView::CreateOpenGLContext(HDC hDC)
{
	// create rendering context associated with windows dc
	m_hRC = ::wglCreateContext(hDC);

	if (m_hRC == NULL)
		return false;	

	// keep it current in the windows lifetime
	// if you do not change it in subclass
	if (::wglMakeCurrent(hDC, m_hRC) == FALSE)
		return false;

	return true;
}


void COpenGLView::OnLButtonDown(UINT nFlags, CPoint point) 
{
	m_ptMouseDownPos = point;
	m_ptMousePrevPos = point;
	m_bLButtonDown = TRUE;

	SetCapture();

	m_pArcball->mouse_down(point.x, point.y);

	CView::OnLButtonDown(nFlags, point);
}


void COpenGLView::OnLButtonUp(UINT nFlags, CPoint point) 
{
	m_bLButtonDown = FALSE;

	ReleaseCapture();

	CView::OnLButtonUp(nFlags, point);
}


void COpenGLView::OnRButtonDown(UINT nFlags, CPoint point) 
{
	m_ptMouseDownPos = point;
	m_ptMousePrevPos = point;
	m_bRButtonDown = TRUE;

	SetCapture();

	CView::OnRButtonDown(nFlags, point);
}


void COpenGLView::OnRButtonUp(UINT nFlags, CPoint point) 
{
	// TODO: Add your message handler code here and/or call default
	m_bRButtonDown = FALSE;

	ReleaseCapture();

	CView::OnRButtonUp(nFlags, point);
}


void COpenGLView::OnMouseMove(UINT nFlags, CPoint point) 
{
	// translation
	if(m_bLButtonDown && m_bRButtonDown)
		ViewTranslation(point, m_ptMousePrevPos);

	// rotation
	else if(m_bLButtonDown) 
		ArcballDragMove(point);

	// scaling
	else if(m_bRButtonDown)
		ViewZoomInOut(point, m_ptMousePrevPos);
	
	m_ptMousePrevPos = point;	
	CView::OnMouseMove(nFlags, point);
}

void COpenGLView::ViewFitAll()
{
	// 粻伴ʵ
	m_fScale = 1.0f;
	m_fTransX = 0.0f;
	m_fTransY = 0.0f;
	m_fTransZ = 0.0f;

	InvalidateRect(NULL, FALSE);
}

void COpenGLView::ArcballDragMove(CPoint point)
{
	m_pArcball->mouse_move(point.x, point.y);
	InvalidateRect(NULL, FALSE);
}

void COpenGLView::ArcballRotation()
{
	m_pArcball->apply_rotation();	
}

void COpenGLView::ViewZoomRect(CRect zoomRect)
{
	zoomRect.NormalizeRect();
	if (zoomRect.IsRectEmpty())
		return;

	CRect screenRect;
	GetClientRect(&screenRect);	

	//////////////////////////////////////////////////////////////////////////
	// new scale 
	float wRatio = float(screenRect.Width()) / zoomRect.Width();
	float hRatio = float(screenRect.Height()) / zoomRect.Height();
	float scaleLast = m_fScale;
	float scale = m_fScale * min(wRatio, hRatio);  // Ŵ = ϴηŴ * ηŴ
	if (scale <= 0)
		m_fScale = 1.0f;              // ŴСڵ
	else
		m_fScale = min(2000, scale);  // Ŵ

	//////////////////////////////////////////////////////////////////////////
	::glPushMatrix();
		::glMatrixMode(GL_MODELVIEW);
		::glLoadIdentity();
		::glTranslatef(m_fTransX, m_fTransY, m_fTransZ); // ƽ
		::glScalef(m_fScale, m_fScale, m_fScale);	// 	
		ArcballRotation();

		::glGetIntegerv(GL_VIEWPORT, m_viewPort);
		::glGetDoublev(GL_MODELVIEW_MATRIX, m_modelViewMat);
		::glGetDoublev(GL_PROJECTION_MATRIX, m_projectionMat);
	::glPopMatrix();

	GLint glY = 0;		
	double realSelCenterX, realSelCenterY, realSelCenterZ;    // ѡĶӦʵռ
	glY = m_viewPort[3] - zoomRect.CenterPoint().y /*- 1*/;
	::gluUnProject(zoomRect.CenterPoint().x, glY, 0.0, m_modelViewMat, m_projectionMat, m_viewPort,
		&realSelCenterX, &realSelCenterY, &realSelCenterZ);

	double realViewCenterX, realViewCenterY, realViewCenterZ; // ӿĶӦʵռ
	glY = m_viewPort[3] - screenRect.CenterPoint().y/* - 1*/;
	::gluUnProject(screenRect.CenterPoint().x, glY, 0.0, m_modelViewMat, m_projectionMat, m_viewPort,
		&realViewCenterX, &realViewCenterY, &realViewCenterZ);

	double dx = realViewCenterX - realSelCenterX;  // ʵάռƫ
	double dy = realViewCenterY - realSelCenterY;
	double dz = realViewCenterZ - realSelCenterZ;

	double x = m_modelViewMat[0] * dx + m_modelViewMat[4] * dy + m_modelViewMat[8]  * dz + m_modelViewMat[12];
	double y = m_modelViewMat[1] * dx + m_modelViewMat[5] * dy + m_modelViewMat[9]  * dz + m_modelViewMat[13];
	double z = m_modelViewMat[2] * dx + m_modelViewMat[6] * dy + m_modelViewMat[10] * dz + m_modelViewMat[14];

	m_fTransX = static_cast<float>(x) *  m_fScale / scaleLast;
	m_fTransY = static_cast<float>(y) *  m_fScale / scaleLast;
	m_fTransZ = static_cast<float>(z) *  m_fScale / scaleLast;

	InvalidateRect(NULL, FALSE);
}

void COpenGLView::ViewZoomInOut(CPoint pos, CPoint prevPos)
{
	if (pos.y - prevPos.y >= 1) 
		m_fScale /= m_fSpeedScale;
	else if (pos.y - prevPos.y <= -1)
		m_fScale *= m_fSpeedScale;

	InvalidateRect(NULL, FALSE);
}

void COpenGLView::ViewTranslation(CPoint pos, CPoint lastPos)
{
	::glGetIntegerv(GL_VIEWPORT, m_viewPort);

	m_fTransX += (float)(pos.x - lastPos.x) / (m_viewPort[3] / 2.0);
	m_fTransY -= (float)(pos.y - lastPos.y) / (m_viewPort[3] / 2.0);

	InvalidateRect(NULL, FALSE);
}

void COpenGLView::EnableBothSidesLight(bool b /* = false */)
{
	m_bEnableBothSidesLight = b;
	if (m_bEnableBothSidesLight)
		::glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE); 
	else
		::glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_FALSE);

	InvalidateRect(NULL, FALSE);
}


// COpenGLView diagnostics

#ifdef _DEBUG
void COpenGLView::AssertValid() const
{
	CView::AssertValid();
}

#ifndef _WIN32_WCE
void COpenGLView::Dump(CDumpContext& dc) const
{
	CView::Dump(dc);
}
#endif
#endif //_DEBUG

