/********************************************************************
* Copyright (c) 2010, Visual Computing Center(VCC), ShenZhen Institute
* of Advanced Technology, Chinese Academy of Sciences. Liangliang Nan, 
* all rights reserved.	
*
*   created:	2006/04/27
*	filename: 	ArcBall.h
*	author:		Liangliang Nan
*   E-mail:     liangliang.nan@gmail.com
*	purpose:	OpenGL wraper class for MFC 
*
*********************************************************************/

//-----------------------------------------------------------------

/*///////////////////////////////////////////////////////////////

This class is a base view-class for simple OpenGL-programming
Applications should subclass this class to get a working view.

This class will draw a yellow sphere on a black background
if your subclass does not redefine any of the functionality
_____________________________________

The class handles the following:
1. Sets up the drawing environment;
2. Do one-time initialization of OpenGL-rendering context;
3. Draws the actual scene;
4. Handles some key messages (see below)
_____________________________

Handling following messages:
1. OnSize(), Redefined in most cases, the viewing pyramid, parallel perspective, front and back clip;
2. OnCreate(), usually not redefined;
3. OnEraseBkgnd(), seldom (never) redefined;
4. OnDestroy(), usually not redefined. Implementations in subclass must include a call to COpenGLView::OnDestroy();
5. PreCreateWindow(), seldom (never) redefined;
6. OnDraw(), seldom (never) redefined.
________________________________

Default funcions
1. InitializeRContext(), virtual function. Allmost allways redefined.
Deals with lighting, back color, shading model etc. 
Things that will usually be done once in the programs lifetime.
2. RenderScene(), virtual function. Allways redefined
_________________________________
*/

#ifndef _OPENGLVIEW_H_
#define _OPENGLVIEW_H_

#include <vector>
#include <gl/gl.h>


class Arcball;

// COpenGLView view
class COpenGLView : public CView
{
	DECLARE_DYNCREATE(COpenGLView)

public:
	//Ľӿ
	virtual void RenderSceneScalable();		                 // drawing code

	void EnableLighting(bool b = true);                     
	void EnableBothSidesLight(bool b = false);               

protected: 
	void ViewTranslation(CPoint pos, CPoint prevPos);      // translation
	void ViewZoomInOut(CPoint pos, CPoint prevPos);		   // zoom in/out
	void ViewZoomRect(CRect zoomRect);					   // zoom the selected rect area to fit the window
	void ViewFitAll();									   // Fit model to the window size. 

	void ArcballDragMove(CPoint point);                      
	void ArcballRotation();

protected:
	bool     m_bEnableLighting;			// Ƿù	
	bool     m_bEnableBothSidesLight;   // Ƿ˫	

	bool     m_bLButtonDown;
	bool     m_bRButtonDown;
	CPoint   m_ptMousePrevPos;          // ǰλ
	CPoint   m_ptMouseDownPos;          // 갴ʱλ

	GLint	 m_viewPort[4];
	double   m_modelViewMat[16];
	double   m_projectionMat[16];

	float    m_fScale;                  // 
	float    m_fTransX;                 // ƽ
	float    m_fTransY;
	float    m_fTransZ;

private:
	void Resize(int cx, int cy) ;  // called by OnSize();
	void InitRenderContext();

	bool SetWindowPixelFormat(HDC hDC); // SetWindowPixelFormat
	bool CreateOpenGLContext(HDC hDC);  // Create an OpenGL rendering context

	void InitGeometry();		// ʼģ͵ʾ

private:
	HGLRC     m_hRC;			// rendering context handle

	float     m_fOrthoLeft;		// ƽͶӰ
	float     m_fOrthoRight;
	float     m_fOrthoTop;
	float     m_fOrthoBottom;
	float     m_fOrthoNear;
	float     m_fOrthoFar;

	Arcball*  m_pArcball;

	float     m_fSpeedScale;

	bool      m_bEnableAntiAliasing;  // Ƿ÷

protected:
	COpenGLView();             // protected constructor used by dynamic creation
	virtual ~COpenGLView();

public:
	virtual void OnDraw(CDC* pDC);    // overridden to draw this view
	virtual BOOL PreCreateWindow(CREATESTRUCT& cs);
#ifdef _DEBUG
	virtual void AssertValid() const;
#ifndef _WIN32_WCE
	virtual void Dump(CDumpContext& dc) const;
#endif
#endif

	// Generated message map functions
protected:
	//{{AFX_MSG(COpenGLView)
	afx_msg int  OnCreate(LPCREATESTRUCT lpCreateStruct);
	afx_msg void OnSize(UINT nType, int cx, int cy);
	afx_msg void OnDestroy();
	afx_msg BOOL OnEraseBkgnd(CDC* pDC);
	afx_msg void OnLButtonDown(UINT nFlags, CPoint point);
	afx_msg void OnLButtonUp(UINT nFlags, CPoint point);
	afx_msg void OnMouseMove(UINT nFlags, CPoint point);
	afx_msg void OnRButtonDown(UINT nFlags, CPoint point);
	afx_msg void OnRButtonUp(UINT nFlags, CPoint point);
	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()
};

#endif

